import logging
import pathlib
import re
from typing import Optional, Tuple, List

import bs4
from w3lib.url import url_query_cleaner
from yarl import URL

from feedsearch_crawler.crawler import Response, Request
from feedsearch_crawler.crawler.lib import parse_href_to_url
from feedsearch_crawler.feed_spider.regexes import (
    feedlike_regex,
    podcast_regex,
    author_regex,
    date_regex,
)

# List of invalid filetypes
invalid_filetypes: List[str] = [
    "jpeg",
    "jpg",
    "png",
    "gif",
    "bmp",
    "mp4",
    "mp3",
    "mkv",
    "md",
    "css",
    "avi",
    "pdf",
    "js",
    "woff",
    "woff2",
    "svg",
    "ttf",
]

# List of strings that are invalid as querystring keys
invalid_querystring_keys: List[str] = ["comment", "comments", "post", "view", "theme"]

# List of strings that indicate a URL is invalid for crawling
invalid_url_contents: List[str] = [
    "wp-includes",
    "wp-content",
    "wp-json",
    "xmlrpc",
    "wp-admin",
    "/amp/",  # Theoretically there could be a feed at an AMP url, but not worth checking.
    "mailto:",
    "//font.",
]

# List of strings that indicate a URL should be low priority
low_priority_urls: List[str] = [
    "/archive/",  # Archives are less likely to contain feeds.
    "/page/",  # Articles pages are less likely to contain feeds.
    "forum",  # Forums are not likely to contain interesting feeds.
    "//cdn.",  # Can't guarantee that someone won't put a feed at a CDN url, so we can't outright ignore it.
    "video",
]

# Link Types that should always be searched for feeds
feed_link_types: List[str] = ["application/json", "rss", "atom", "rdf"]


logger = logging.getLogger(__name__)


class LinkFilter:
    def __init__(self, response: Response, request: Request, full_crawl: bool = False):
        self.response = response
        self.request = request
        self.full_crawl = full_crawl

    def should_follow_link(self, link: bs4.Tag) -> Optional[Tuple[URL, int]]:
        """
        Check that the link should be followed if it may contain feed information.

        :param link: Link tag
        :return: boolean
        """
        href: str = link.get("href")
        link_type: str = link.get("type")

        url: URL = parse_href_to_url(href)
        if not url:
            return None

        # If the link may have a valid feed type then follow it regardless of the url text.
        if (
            link_type
            and any(map(link_type.lower().count, feed_link_types))
            and "json+oembed" not in link_type
        ):
            # A link with a possible feed type has the highest priority after callbacks.
            return url, 2

        is_feedlike_href: bool = self.is_href_matching(str(url), feedlike_regex)
        is_feedlike_querystring: bool = self.is_querystring_matching(
            url, feedlike_regex
        )

        is_podcast_href: bool = self.is_href_matching(str(url), podcast_regex)
        is_podcast_querystring: bool = self.is_querystring_matching(url, podcast_regex)

        is_feedlike_url = is_feedlike_querystring or is_feedlike_href
        is_podcast_url = is_podcast_href or is_podcast_querystring

        if not self.full_crawl and not is_feedlike_url and not is_podcast_url:
            return

        # This check is deprecated, as it has been moved to the spider to prevent the crawling of any links
        # from responses that are not the same as the original domain
        #
        # is_one_jump: bool = self.is_one_jump_from_original_domain(url, self.response)
        # if not is_one_jump:
        #     return

        has_author_info: bool = self.is_href_matching(href, author_regex)
        is_low_priority: bool = self.is_low_priority(href)

        priority: int = Request.priority
        # A low priority url should be fetched last.
        if is_low_priority:
            priority = Request.priority + 2
        # Podcast pages are lower priority than authors or feeds.
        if is_podcast_url:
            priority = 5
        # Potential author info has a medium priority.
        if has_author_info:
            priority = 4
        # A feedlike url has high priority.
        if is_feedlike_url:
            priority = 3

        # Validate the actual URL string.
        follow = (
            # is_one_jump
            not self.has_invalid_contents(href)
            and self.is_valid_filetype(href)
            and not self.has_invalid_querystring(url)
        )
        # If full_crawl then follow all valid URLs regardless of the feedlike quality of the URL.
        # Otherwise only follow URLs if they look like they might contain feed information.
        if follow and (self.full_crawl or is_feedlike_url or is_podcast_href):

            # Remove the querystring unless it may point to a feed.
            if not is_feedlike_querystring:
                url = url.with_query(None)

            return url, priority

    @staticmethod
    def is_one_jump_from_original_domain(url: URL, response: Response) -> bool:
        """
        Check that the current URL is only one response away from the originally queried domain.

        We want to be able to follow potential feed links that point to a different domain than
        the originally queried domain, but not to follow any deeper than that.

        Sub-domains of the original domain are ok.

        i.e: the following are ok
            "test.com" -> "feedhost.com"
            "test.com/feeds" -> "example.com/feed.xml"
            "test.com" -> "feeds.test.com"

        not ok:
            "test.com" -> "feedhost.com" (we stop here) -> "feedhost.com/feeds"

        :param url: URL object or string
        :param response: Response object
        :return: boolean
        """

        # This is the first Response in the chain
        if len(response.history) < 2:
            return True

        # The URL is relative, so on the same domain
        if not url.is_absolute():
            return True

        # URL is same domain
        if url.host == response.history[0].host:
            return True

        # URL is sub-domain
        if response.history[0].host in url.host:
            return True

        # URL domain and current Response domain are different from original domain
        if (
            response.history[-1].host != response.history[0].host
            and url.host != response.history[0].host
        ):
            return False

        return True

    @staticmethod
    def is_valid_filetype(url: str) -> bool:
        """
        Check if url string has an invalid filetype extension.

        :param url: URL string
        :return: boolean
        """
        # if file_regex.search(url.strip()):
        #     return False
        # return True
        suffix = pathlib.Path(url_query_cleaner(url)).suffix.strip(".").lower()
        if suffix in invalid_filetypes:
            return False
        return True

    @staticmethod
    def has_invalid_querystring(url: URL) -> bool:
        """
        Check if URL querystring contains invalid keys.

        :param url: URL object
        :return: boolean
        """
        return any(key in url.query for key in invalid_querystring_keys)

    @staticmethod
    def is_href_matching(url_string: str, regex: re) -> bool:
        """
        Check if the regex has any match in the url string.

        :param url_string: URL as string
        :param regex: Regex used to search URL
        :return: boolean
        """
        if regex.search(url_query_cleaner(url_string)):
            return True
        return False

    @staticmethod
    def is_querystring_matching(url: URL, regex: re) -> bool:
        """
        Check if the regex has any match in the URL query parameters.

        :param url: URL object
        :param regex: Regex used to search query
        :return: boolean
        """
        for key in url.query:
            if regex.search(key):
                return True
        return False

    @staticmethod
    def has_invalid_contents(string: str) -> bool:
        """
        Ignore any string containing the following strings.

        :param string: String to check
        :return: boolean
        """
        return any(value in string.lower() for value in invalid_url_contents)

    @staticmethod
    def is_low_priority(url_string: str) -> bool:
        """
        Check if the url contains any strings that indicate the url should be low priority.

        :param url_string: URL string
        :return: boolean
        """
        if any(value in url_string.lower() for value in low_priority_urls):
            return True

        # Search for dates in url, this generally indicates an article page.
        if date_regex.search(url_string):
            return True
        return False

    @staticmethod
    def is_subdomain_matching(url: URL, regex: re) -> bool:
        """
        Check if the url subdomain matches the regex

        :param url: URL object
        :param regex: regex object
        :return: boolean
        """
        if not url.host:
            return False

        split = url.host.split(".")
        if len(split) <= 2:
            return False

        sub_domains = ".".join(split[:-2])
        if regex.search(sub_domains):
            return True
        return False



import base64
import logging
from types import AsyncGeneratorType
from typing import Union, Any, List, Set

import bs4
from yarl import URL

from feedsearch_crawler.crawler import Crawler, Item, Request, Response
from feedsearch_crawler.crawler.lib import parse_href_to_url
from feedsearch_crawler.feed_spider.dupefilter import NoQueryDupeFilter
from feedsearch_crawler.feed_spider.favicon import Favicon
from feedsearch_crawler.feed_spider.feed_info import FeedInfo
from feedsearch_crawler.feed_spider.feed_info_parser import FeedInfoParser
from feedsearch_crawler.feed_spider.lib import ParseTypes
from feedsearch_crawler.feed_spider.link_filter import LinkFilter
from feedsearch_crawler.feed_spider.regexes import rss_regex
from feedsearch_crawler.feed_spider.site_meta import SiteMeta
from feedsearch_crawler.feed_spider.site_meta_parser import SiteMetaParser

logger = logging.getLogger(__name__)


class FeedsearchSpider(Crawler):
    duplicate_filter_class = NoQueryDupeFilter
    htmlparser = "html.parser"
    favicon_data_uri = True
    try_urls: Union[List[str], bool] = False
    full_crawl: bool = False
    crawl_hosts: bool = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.site_meta_processor = SiteMetaParser(self)
        self.feed_info_parser = FeedInfoParser(self)
        self.site_metas = set()
        self.favicons = dict()
        self.feeds_seen = dict()
        self.post_crawl_callback = self.populate_feed_site_meta
        if "try_urls" in kwargs:
            self.try_urls = kwargs["try_urls"]
        if "favicon_data_uri" in kwargs:
            self.favicon_data_uri = kwargs["favicon_data_uri"]
        if "full_crawl" in kwargs:
            self.full_crawl = kwargs["full_crawl"]
        if "crawl_hosts" in kwargs:
            self.crawl_hosts = kwargs["crawl_hosts"]

    async def parse(self, request: Request, response: Response) -> AsyncGeneratorType:
        """
        Parse a Response for feeds or site metadata.

        :param request: Request
        :param response: Response
        :return: AsyncGenerator yielding Items, Requests, or iterative AsyncGenerators
        """

        # If the Response is not OK then there's no data to parse.
        if not response.ok:
            return

        # If the Response contains JSON then attempt to parse it as a JsonFeed.
        if response.json:
            if "version" and "jsonfeed" and "feed_url" in response.json:
                yield self.feed_info_parser.parse_item(
                    request, response, parse_type=ParseTypes.JSON
                )
                return

        if not isinstance(response.text, str):
            logger.debug("No text in %s", response)
            return

        yield self.parse_site_meta(request, response)

        # Restrict the RSS check to the first 1000 characters, otherwise it's almost definitely not an actual feed.
        if rss_regex.search(response.text, endpos=1000):
            yield self.feed_info_parser.parse_item(
                request, response, parse_type=ParseTypes.XML
            )
            return

        # Don't waste time trying to parse and follow urls if the max depth is already reached.
        if response.is_max_depth_reached(self.max_depth):
            logger.debug("Max depth %d reached: %s", self.max_depth, response)
            return

        # Make sure the Response XML has been parsed if it exists.
        soup = await response.xml
        if not soup:
            return

        # Don't crawl links from pages that are not from the original domain
        if not response.is_original_domain():
            return

        link_filter = LinkFilter(
            request=request, response=response, full_crawl=self.full_crawl
        )

        # Find all links in the Response.
        links = soup.find_all(self.tag_has_href)
        for link in links:
            # Check each href for validity and queue priority.
            values = link_filter.should_follow_link(link)
            if values:
                url, priority = values
                yield await self.follow(
                    url, self.parse, response, priority=priority, allow_domain=True
                )

    async def parse_site_meta(
        self, request: Request, response: Response
    ) -> AsyncGeneratorType:
        """
        Parses site metadata if the returned URL is a site origin URL.

        If the returned url is an origin url, or the request url is an origin url (and there may have been a redirect)
        then parse the site meta.

        :param request: Request
        :param response: Response
        :return: AsyncGenerator yielding SiteMeta items
        """
        url_origin = response.url.origin()
        request_url_origin = request.url.origin()

        if response.url == url_origin or request.url == request_url_origin:
            yield self.site_meta_processor.parse_item(request, response)

    async def parse_xml(self, response_text: str) -> Any:
        """
        Parse Response text as XML.
        Used to allow implementations to provide their own XML parser.

        :param response_text: Response text as string.
        :return: None
        """
        return bs4.BeautifulSoup(response_text, self.htmlparser)

    async def process_item(self, item: Item) -> None:
        """
        Process parsed items.

        :param item: Item object
        :return: None
        """
        if isinstance(item, FeedInfo):
            self.items.add(item)
        elif isinstance(item, SiteMeta):
            self.site_metas.add(item)
        elif isinstance(item, Favicon):
            self.add_favicon(item)

    def add_favicon(self, favicon: Favicon) -> None:
        """
        Add a favicon to the spider's favicon dictionary.

        :param favicon: Favicon object
        """
        existing: Favicon = self.favicons.get(favicon.url)
        if existing and existing.data_uri and not favicon.data_uri:
            return
        self.favicons[favicon.url] = favicon

    # noinspection PyPep8
    async def populate_feed_site_meta(self) -> None:
        """
        Populate FeedInfo site information with data from the relevant SiteMeta item
        """
        for feed in self.items:
            # Check each SiteMeta for a url host match
            site_meta = next(
                (x for x in self.site_metas if x.host in feed.url.host), None
            )
            if site_meta:
                feed.site_url = site_meta.url
                feed.site_name = site_meta.site_name

            # Populate favicon directly if available
            if feed.favicon:
                favicon = self.favicons.get(feed.favicon)
                if favicon:
                    feed.favicon_data_uri = favicon.data_uri
                    feed.favicon = favicon.resp_url if favicon.resp_url else favicon.url

            # If a favicon hasn't been found yet or there is no data_uri then try and find a suitable favicon
            if not feed.favicon or (
                self.favicon_data_uri and not feed.favicon_data_uri
            ):
                feed_host = feed.url.host
                favicons = list(
                    x
                    for x in self.favicons.values()
                    if x.matches_host(feed_host, self.favicon_data_uri)
                )

                if favicons:
                    favicon = min(favicons, key=lambda x: x.priority)

                    feed.favicon_data_uri = favicon.data_uri
                    feed.favicon = favicon.resp_url if favicon.resp_url else favicon.url

    # noinspection PyUnusedLocal
    async def parse_favicon_data_uri(
        self, request: Request, response: Response, favicon: Favicon
    ) -> None:
        """
        Create a data uri from a favicon image.

        :param request: Request
        :param response: Response
        :param favicon: Favicon object
        :return: None
        """
        if not response.ok or not response.data or not isinstance(response.data, bytes):
            return

        def is_png(data: bytes) -> bool:
            return data[:8] in bytes.fromhex("89 50 4E 47 0D 0A 1A 0A")

        def is_ico(data: bytes) -> bool:
            return data[:4] in bytes.fromhex("00 00 01 00")

        try:
            if not is_png(response.data) and not is_ico(response.data):
                logger.debug("Response data is not a valid image type: %s", response)
                return
        except Exception as e:
            logger.exception("Failure validation image type: %s: %s", response, e)

        try:
            encoded = base64.b64encode(response.data)
            uri = "data:image/png;base64," + encoded.decode(response.encoding)
            favicon.resp_url = response.url
            favicon.data_uri = uri
            self.add_favicon(favicon)
        except Exception as e:
            logger.exception("Failure encoding image: %s: %s", response, e)

    def create_start_urls(self, urls: List[Union[URL, str]]) -> List[URL]:
        """
        Create the start URLs for the crawl from an initial URL. May be overridden.

        :param urls: Initial URLs
        """
        crawl_start_urls: Set[URL] = set()

        for url in urls + self.start_urls:
            if isinstance(url, str):
                if "//" not in url:
                    url = f"//{url}"
                url = parse_href_to_url(url)
                if not url:
                    continue

            if url.scheme.lower() not in ["http", "https"]:
                url = url.with_scheme("http")

            crawl_start_urls.add(url)

        origins = set(url.origin() for url in crawl_start_urls)

        if self.try_urls:
            # Common paths for feeds.
            suffixes = {
                "index.xml",
                "atom.xml",
                "feeds",
                "feeds/default",
                "feed",
                "feed/default",
                "feeds/posts/default",
                "?feed=rss",
                "?feed=atom",
                "?feed=rss2",
                "?feed=rdf",
                "rss",
                "atom",
                "rdf",
                "index.rss",
                "index.rdf",
                "index.atom",
                "data/rss",
                "rss.xml",
                "index.json",
                "about",
                "about/feeds",
                "rss-feeds",
            }

            for origin in origins:
                if isinstance(self.try_urls, list):
                    crawl_start_urls.update(
                        origin.join(URL(suffix)) for suffix in self.try_urls
                    )
                else:
                    crawl_start_urls.update(
                        origin.join(URL(suffix)) for suffix in suffixes
                    )

        # Crawl the origin urls of the start urls for Site metadata.
        if self.crawl_hosts:
            crawl_start_urls.update(origins)

        return list(crawl_start_urls)

    @staticmethod
    def tag_has_href(tag: bs4.Tag) -> bool:
        """
        Find all tags that contain links.

        :param tag: XML tag
        :return: boolean
        """
        return tag.has_attr("href")



from yarl import URL

from feedsearch_crawler.crawler import Item


class Favicon(Item):
    url: URL = None
    priority: int = 0
    rel: str = ""
    data_uri: str = ""
    resp_url: URL = None
    site_host: str = ""

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.url == other.url

    def __hash__(self):
        return hash(self.url)

    def __repr__(self):
        return f"{self.__class__.__name__}({str(self.url)})"

    def matches_host(self, host: str, requires_data_uri: bool = False) -> bool:
        """
        Check that the Favicon site_host is a match for the host.

        :param host: domain host url string
        :param requires_data_uri: Whether the Favicon is required to have a data_uri
        :return: bool
        """
        return (
            self.url
            and self.site_host
            and self.site_host in host
            and (self.data_uri if requires_data_uri else True)
        )



import asyncio
import logging
from xml.etree import ElementTree
from typing import List, Union

from yarl import URL

from feedsearch_crawler.feed_spider import FeedsearchSpider, FeedInfo

logging.getLogger(__name__).addHandler(logging.NullHandler())

name = "Feedsearch Crawler"


def search(
    url: Union[URL, str, List[Union[URL, str]]],
    try_urls: Union[List[str], bool] = False,
    *args,
    **kwargs
) -> List[FeedInfo]:
    """
    Search for feeds at a URL.

    :param url: URL or list of URLs to search
    :param try_urls: Tries different paths that may contain feeds.
    :return: List of FeedInfo objects
    """
    results = asyncio.run(search_async(url, try_urls=try_urls, *args, **kwargs))
    return results


async def search_async(
    url: Union[URL, str, List[Union[URL, str]]],
    try_urls: Union[List[str], bool] = False,
    *args,
    **kwargs
) -> List[FeedInfo]:
    """
    Search asynchronously for feeds at a URL.

    :param url: URL or list of URLs to search
    :param try_urls: Tries different paths that may contain feeds.
    :return: List of FeedInfo objects
    """
    crawler = FeedsearchSpider(try_urls=try_urls, *args, **kwargs)
    await crawler.crawl(url)

    return sort_urls(list(crawler.items))


def sort_urls(feeds: List[FeedInfo]) -> List[FeedInfo]:
    """
    Sort list of feeds based on Url score

    :param feeds: List of FeedInfo objects
    :return: List of FeedInfo objects sorted by score
    """
    feeds = [f for f in feeds if isinstance(f, FeedInfo)]
    sorted_urls = sorted(list(set(feeds)), key=lambda x: x.score, reverse=True)
    return sorted_urls


def output_opml(feeds: List[FeedInfo]) -> bytes:
    """
    Return feeds as a subscriptionlist OPML file.
    http://dev.opml.org/spec2.html#subscriptionLists

    :param feeds: List of FeedInfo objects
    :return: OPML file as XML bytestring
    """
    root = ElementTree.Element("opml", version="2.0")
    head = ElementTree.SubElement(root, "head")
    title = ElementTree.SubElement(head, "title")
    title.text = "Feeds"
    body = ElementTree.SubElement(root, "body")

    for feed in feeds:
        if not feed.url:
            continue

        fe = ElementTree.SubElement(body, "outline", type="rss", xmlUrl=str(feed.url))

        if feed.title:
            fe.set("text", feed.title)
            fe.set("title", feed.title)
        if feed.site_url:
            fe.set("htmlUrl", str(feed.site_url))
        if feed.description:
            fe.set("description", feed.description)
        if feed.version:
            fe.set("version", feed.version)

    return ElementTree.tostring(root, encoding="utf8", method="xml")



from abc import ABC


class Item(ABC):
    ignore_item = False

    def __init__(self, **kwargs):
        for k in kwargs.keys():
            if hasattr(self, k):
                self.__setattr__(k, kwargs[k])



from w3lib.url import url_query_cleaner, canonicalize_url
from yarl import URL

from feedsearch_crawler.crawler import DuplicateFilter


class NoQueryDupeFilter(DuplicateFilter):
    valid_keys = ["feedformat", "feed", "rss", "atom", "jsonfeed", "format", "podcast"]

    def parse_url(self, url: URL) -> str:
        # Keep the query strings if they might be feed strings.
        # Wikipedia for example uses query strings to differentiate feeds.
        if any(key in url.query for key in self.valid_keys):
            return canonicalize_url(str(url))

        # Canonicalizing the URL is about 4x slower, but worth it to prevent duplicate requests.
        return canonicalize_url(url_query_cleaner(str(url)))



import cgi
from datetime import datetime
from typing import Union, List

from dateutil import tz, parser
from yarl import URL


class ParseTypes:
    JSON = "json"
    XML = "xml"


def get_site_root(url: Union[str, URL]) -> str:
    """
    Find the root domain of a url
    """
    if isinstance(url, URL):
        return url.host
    return URL(url).host


def create_allowed_domains(url: Union[str, URL]) -> List[str]:
    if isinstance(url, URL):
        return [url.host]
    return [URL(url).host]


def parse_header_links(value):
    """
    Return a list of Dicts of parsed link headers proxies.
    i.e. Link: <http:/.../front.jpeg>; rel=front; type="image/jpeg",
    <http://.../back.jpeg>; rel=back;type="image/jpeg"

    :param value: HTTP Link header to parse
    :return: List of Dicts
    """

    links = []

    replace_chars = " '\""

    for val in value.split(","):
        try:
            url, params = val.split(";", 1)
        except ValueError:
            url, params = val, ""

        link = {"url": url.strip("<> '\"")}

        for param in params.split(";"):
            try:
                key, value = param.split("=")
            except ValueError:
                break

            link[key.strip(replace_chars)] = value.strip(replace_chars)

        links.append(link)

    return links


def force_utc(dt: datetime) -> datetime:
    """
    Change a datetime to UTC, and convert naive datetimes to tz-aware UTC.

    :param dt: datetime to change to UTC
    :return: tz-aware UTC datetime
    """
    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=tz.tzutc())
    return dt.astimezone(tz.tzutc())


def datestring_to_utc_datetime(date_string: str) -> datetime:
    """
    Convert a date string to a tz-aware UTC datetime.

    :param date_string: A datetime as a string in almost any format.
    :return: tz-aware UTC datetime
    """
    dt = parser.parse(date_string)
    return force_utc(dt)


def create_content_type(parse_type: str, encoding: str, content_type: str) -> str:
    """
    Create the actual content type of the feed.

    :param parse_type: How the feed is being parsed. XML or JSON
    :param encoding: Charset encoding of the response
    :param content_type: Content-Type header string of the response
    :return: Content-Type string
    """
    ctype, pdict = cgi.parse_header(content_type)

    if parse_type == ParseTypes.JSON and ParseTypes.JSON not in ctype.lower():
        ctype = "application/json"
    elif parse_type == ParseTypes.XML and ParseTypes.XML not in ctype.lower():
        ctype = "application/xml"

    return f"{ctype}; charset={encoding}".lower()



import re

# Regex to check if possible RSS data.
rss_regex = re.compile("(<rss|<rdf|<feed)", re.IGNORECASE)

# Regex to check that a feed-like string is a whole word to help rule out false positives.
feedlike_regex = re.compile(
    "\\b(rss|feeds?|atom|json|xml|rdf|blogs?)\\b", re.IGNORECASE
)

# Regex to check that a podcast string is a whole word.
podcast_regex = re.compile("\\b(podcasts?)\\b", re.IGNORECASE)

# Regex to check if the URL might contain author information.
author_regex = re.compile(
    "(authors?|journalists?|writers?|contributors?)", re.IGNORECASE
)

# Regex to check URL string for invalid file types.
file_regex = re.compile(
    ".(jpe?g|png|gif|bmp|mp4|mp3|mkv|md|css|avi|pdf|js|woff2?|svg|ttf|zip)/?$",
    re.IGNORECASE,
)

# Regex to match year and month in URLs, e.g. /2019/07/
date_regex = re.compile("/(\\d{4}/\\d{2})/")



import uuid
from typing import List, Dict, Any, Optional

from yarl import URL

from feedsearch_crawler.crawler.lib import is_same_domain


class Response:
    _xml = None

    def __init__(
        self,
        url: URL,
        method: str,
        encoding: str = "",
        text: str = "",
        json: Dict = None,
        data: bytes = b"",
        history: List[URL] = None,
        headers=None,
        status_code: int = -1,
        cookies=None,
        xml_parser=None,
        redirect_history=None,
        content_length: int = 0,
        meta: Dict = None,
    ):
        self.url = url
        self.encoding = encoding
        self.method = method
        self.text = text
        self.json = json
        self.data = data
        self.history = history or []
        self.headers = headers or {}
        self.status_code = status_code
        self.cookies = cookies
        self.id = uuid.uuid4()
        self._xml_parser = xml_parser
        self.redirect_history = redirect_history
        self.content_length = content_length
        self.meta = meta
        self.origin: URL = url.origin()

    @property
    def ok(self) -> bool:
        return self.status_code == 0 or 200 <= self.status_code <= 299

    @property
    def domain(self) -> str:
        return self.url.host

    @property
    def scheme(self) -> str:
        return self.url.scheme

    @property
    def previous_domain(self) -> str:
        if not self.history:
            return ""
        return self.history[-1].host

    @property
    def originator_url(self) -> Optional[URL]:
        if not self.history or len(self.history) == 1:
            return None
        return self.history[-2]

    @property
    async def xml(self) -> Any:
        if self._xml:
            return self._xml

        if not self._xml_parser:
            return None

        if not self.text and self.data and self.encoding:
            self.text = self.data.decode(self.encoding)

        self._xml = await self._xml_parser(self.text)
        return self._xml

    def is_max_depth_reached(self, max_depth: int) -> bool:
        """
        Check if the max response depth has been reached.

        :param max_depth: Max length of response history
        :return: boolean
        """
        if max_depth and len(self.history) >= max_depth:
            return True
        return False

    def is_original_domain(self) -> bool:
        """
        Check if this response is still at the original domain in the response chain.

        :return: boolean
        """
        # This is the first Response in the chain
        if len(self.history) < 2:
            return True
        # URL is same domain or sub-domain
        if is_same_domain(self.history[0].host, self.url.host):
            return True

        return False

    def __repr__(self):
        return f"{self.__class__.__name__}({str(self.url)})"



import asyncio
import copy
import inspect
import logging
from abc import ABC, abstractmethod
from collections import OrderedDict
from fnmatch import fnmatch
from statistics import harmonic_mean, median
from types import AsyncGeneratorType
from typing import List, Any, Dict, Set
from typing import Union

import aiohttp
import time
from aiohttp import ClientTimeout
from yarl import URL

from feedsearch_crawler.crawler.duplicatefilter import DuplicateFilter
from feedsearch_crawler.crawler.item import Item
from feedsearch_crawler.crawler.lib import (
    coerce_url,
    ignore_aiohttp_ssl_error,
    Stats,
    CallbackResult,
    CrawlerPriorityQueue,
    parse_href_to_url,
)
from feedsearch_crawler.crawler.queueable import Queueable
from feedsearch_crawler.crawler.request import Request
from feedsearch_crawler.crawler.response import Response
from feedsearch_crawler.crawler.trace import add_trace_config

try:
    import uvloop

    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
except ImportError:
    uvloop = None
    pass


logger = logging.getLogger(__name__)


class Crawler(ABC):

    # Class Name of the Duplicate Filter.
    # May be overridden to use different Duplicate Filter.
    # Not an instantiation of the class.
    duplicate_filter_class = DuplicateFilter

    # Callback to be run after all workers are finished.
    post_crawl_callback = None

    # URLs to start the crawl.
    start_urls = []
    # Domain patterns that are allowed to be crawled.
    allowed_domains = []

    # Max number of concurrent http requests.
    concurrency: int = 10
    # Max size of incoming http response content.
    max_content_length = 1024 * 1024 * 10
    # Max crawl depth. i.e. The max length of the response history.
    max_depth: int = 4
    # Max callback recursion depth, to prevent accidental infinite recursion from AsyncGenerators.
    max_callback_recursion: int = 10
    # Time in seconds to delay each HTTP request.
    delay: float = 0

    # List of worker tasks.
    _workers = []

    # ClientSession for requests. Created on Crawl start.
    _session: aiohttp.ClientSession
    # Task queue for Requests. Created on Crawl start.
    _request_queue: CrawlerPriorityQueue
    # Semaphore for controlling HTTP Request concurrency.
    _semaphore: asyncio.Semaphore

    def __init__(
        self,
        start_urls: List[str] = None,
        allowed_domains: List[str] = None,
        concurrency: int = 10,
        total_timeout: Union[float, ClientTimeout] = 30,
        request_timeout: Union[float, ClientTimeout] = 5,
        user_agent: str = "",
        max_content_length: int = 1024 * 1024 * 10,
        max_depth: int = 10,
        headers: dict = None,
        allowed_schemes: List[str] = None,
        delay: float = 0.5,
        max_retries: int = 3,
        ssl: bool = False,
        trace: bool = False,
        *args,
        **kwargs,
    ):
        """
        Base class for a WebCrawler implementation.

        :param allowed_schemes: List of strings of allowed Request URI schemes. e.g. ["http", "https"]
        :param start_urls: List of initial URLs to crawl.
        :param allowed_domains: List of domain patterns that are allowed. Uses Unix shell-style wildcards.
        :param concurrency: Max number of workers and of concurrent HTTP requests.
        :param total_timeout: Total aiohttp ClientSession timeout in seconds.
            Crawl will end if this timeout is triggered.
        :param request_timeout: Total timeout in seconds for each individual HTTP request.
        :param user_agent: Default User-Agent for HTTP requests.
        :param max_content_length: Max size in bytes of incoming http response content.
        :param max_depth: Max crawl depth. i.e. The max length of the response history.
        :param headers: Default HTTP headers to be included in each request.
        :param delay: Time in seconds to delay each HTTP request.
        :param max_retries: Maximum number of retries for each failed HTTP request.
        :param ssl: Enables strict SSL checking.
        :param trace: Enables aiohttp trace debugging.
        :param args: Additional positional arguments for subclasses.
        :param kwargs: Additional keyword arguments for subclasses.
        """
        self.start_urls = start_urls or []
        self.allowed_domains = allowed_domains or []

        self.concurrency = concurrency

        if not isinstance(total_timeout, ClientTimeout):
            total_timeout = aiohttp.ClientTimeout(total=total_timeout)
        if not isinstance(request_timeout, ClientTimeout):
            request_timeout = aiohttp.ClientTimeout(total=request_timeout)

        self.total_timeout: ClientTimeout = total_timeout
        self.request_timeout: ClientTimeout = request_timeout

        self.max_content_length = max_content_length
        self.max_depth = max_depth

        self.user_agent = user_agent or (
            "Mozilla/5.0 (compatible; Feedsearch-Crawler; +https://pypi.org/project/feedsearch-crawler)"
        )

        self.headers = {"User-Agent": self.user_agent, "Upgrade-Insecure-Requests": "1"}

        if headers:
            self.headers = {**self.headers, **headers}

        self.allowed_schemes = allowed_schemes
        self.delay = delay
        self.max_retries = max_retries
        self._ssl = ssl
        self._trace = trace

        # Default set for parsed items.
        self.items: set = set()

        # URL Duplicate Filter instance.
        self._duplicate_filter = self.duplicate_filter_class()

        # List of total durations in Milliseconds for the total handling time of all Requests.
        self._stats_request_durations = []
        # List of total duration in Milliseconds of all HTTP requests.
        self._stats_request_latencies = []
        # List of Content Length in bytes of all Responses.
        self._stats_response_content_lengths = []
        # List of time in Milliseconds that each item spend on the queue.
        self._stats_queue_wait_times = []
        # List of the size of the queue each time an item was popped off the queue.
        self._stats_queue_sizes = []

        # Initialise Crawl Statistics.
        self.stats: dict = {
            Stats.REQUESTS_QUEUED: 0,
            Stats.REQUESTS_SUCCESSFUL: 0,
            Stats.REQUESTS_FAILED: 0,
            Stats.CONTENT_LENGTH_TOTAL: 0,
            Stats.CONTENT_LENGTH_AVG: 0,
            Stats.CONTENT_LENGTH_MIN: 0,
            Stats.CONTENT_LENGTH_MAX: 0,
            Stats.CONTENT_LENGTH_MEDIAN: 0,
            Stats.ITEMS_PROCESSED: 0,
            Stats.URLS_SEEN: 0,
            Stats.REQUESTS_DURATION_AVG: 0,
            Stats.REQUESTS_DURATION_MAX: 0,
            Stats.REQUESTS_DURATION_MIN: 0,
            Stats.REQUESTS_DURATION_TOTAL: 0,
            Stats.REQUESTS_DURATION_MEDIAN: 0,
            Stats.TOTAL_DURATION: 0,
            Stats.STATUS_CODES: {},
            Stats.QUEUE_WAIT_MAX: 0,
            Stats.QUEUE_WAIT_MIN: 0,
            Stats.QUEUE_WAIT_AVG: 0,
            Stats.QUEUE_WAIT_MEDIAN: 0,
            Stats.QUEUE_SIZE_MAX: 0,
            Stats.QUEUE_SIZE_AVG: 0,
            Stats.QUEUE_SIZE_MEDIAN: 0,
            Stats.QUEUED_TOTAL: 0,
            Stats.REQUESTS_RETRIED: 0,
        }

    async def _handle_request(self, request: Request) -> None:
        """
        Handle fetching of Requests and processing of Request callbacks.

        :param request: Request
        :return: None
        """
        try:
            if request.has_run and not request.should_retry:
                logger.warning("%s has already run", request)
                return

            start = time.perf_counter()

            # Fetch the request and run its callback
            # results, response = await request.fetch_callback(self._semaphore)
            results, response = await request.fetch_callback()

            dur = int((time.perf_counter() - start) * 1000)
            self._stats_request_durations.append(dur)
            self._stats_request_latencies.append(request.req_latency)
            logger.debug(
                "Fetched: url=%s dur=%dms latency=%dms read=%dms status=%s prev=%s",
                response.url,
                dur,
                request.req_latency,
                request.content_read,
                response.status_code,
                response.originator_url,
            )

            if response.ok:
                self.stats[Stats.REQUESTS_SUCCESSFUL] += 1
            else:
                self.stats[Stats.REQUESTS_FAILED] += 1

            if response.status_code in self.stats[Stats.STATUS_CODES]:
                self.stats[Stats.STATUS_CODES][response.status_code] += 1
            else:
                self.stats[Stats.STATUS_CODES][response.status_code] = 1

            self._stats_response_content_lengths.append(response.content_length)

            # Mark the Response URL as seen in the duplicate filter, as it may be different from the Request URL
            # due to redirects.
            await self._duplicate_filter.url_seen(response.url, response.method)

            # Add callback results to the queue for processing.
            if results:
                self._put_queue(CallbackResult(results, 0))

            # Add Request back to the queue for retrying.
            if request.should_retry:
                self.stats[Stats.REQUESTS_RETRIED] += 1
                self._put_queue(request)

        except asyncio.CancelledError as e:
            logger.debug("Cancelled: %s, %s", request, e)
        except Exception as e:
            logger.exception("Exception during %s: %s", request, e)
        finally:
            return

    async def _process_request_callback_result(
        self, result: Any, callback_recursion: int = 0
    ) -> None:
        """
        Process the Request callback result depending on the result type.
        Request callbacks may contain nested iterators.

        :param result: Callback Result. May be an CallbackResult class, AsyncGenerator, Coroutine, Request, or Item.
        :param callback_recursion: Incremented counter to limit this method's recursion.
        :return: None
        """
        if callback_recursion >= self.max_callback_recursion:
            logger.warning(
                "Max callback recursion of %d reached", self.max_callback_recursion
            )
            return

        try:
            # If a CallbackResult class is passed, process the result values from within the class.
            if isinstance(result, CallbackResult):
                await self._process_request_callback_result(
                    result.result, result.callback_recursion
                )
            # For async generators, put each value back on the queue for processing.
            # This will happen recursively until the end of the recursion chain or max_callback_recursion is reached.
            elif inspect.isasyncgen(result):
                async for value in result:
                    if value:
                        self._put_queue(CallbackResult(value, callback_recursion + 1))
            # For coroutines, await the result then put the value back on the queue for further processing.
            elif inspect.iscoroutine(result):
                value = await result
                self._put_queue(CallbackResult(value, callback_recursion + 1))
            # Requests are put onto the queue to be fetched.
            elif isinstance(result, Request):
                self._process_request(result)

            # Items are handled by the implementing Class.
            elif isinstance(result, Item):
                await self.process_item(result)
                self.stats[Stats.ITEMS_PROCESSED] += 1
        except Exception as e:
            logger.exception(e)

    def _process_request(self, request: Request) -> None:
        """
        Process a Request onto the Request Queue.

        :param request: HTTP Request
        :return: None
        """
        if not request:
            return

        self.stats[Stats.REQUESTS_QUEUED] += 1
        logger.debug("Queue Add: %s", request)
        # Add the Request to the queue for processing.
        self._put_queue(request)

    def is_allowed_domain(self, url: URL) -> bool:
        """
        Check that the URL host is in the list of allowed domain patterns.
        Domain patterns are Unix shell-style wildcards.
        https://docs.python.org/3/library/fnmatch.html

        :param url: URL object
        :return: boolean
        """
        if not self.allowed_domains:
            return True

        try:
            if not url or not url.host:
                return False
            host = url.host
            for domain_pattern in self.allowed_domains:
                if fnmatch(host, domain_pattern):
                    return True
        except Exception as e:
            logger.warning(e)
        return False

    async def follow(
        self,
        url: Union[str, URL],
        callback=None,
        response: Response = None,
        method: str = "GET",
        delay: Union[float, None] = None,
        priority: int = 0,
        allow_domain: bool = False,
        cb_kwargs: Dict = None,
        max_content_length: int = None,
        timeout: float = None,
        retries: int = None,
        **kwargs,
    ) -> Union[Request, None]:
        """
        Follow a URL by creating an HTTP Request.

        If the URL is not absolute then it is joined with the previous Response URL.
        The previous Response history is copied to the Request.

        Before a Request is followed, first check that the Request URL has not already been seen,
        that the max URL depth has not been reached, and that the URI scheme is allowed.

        These checks are performed before the Request is created so that we don't yield multiple requests
        to the same URL to the queue for further processing. We want to stop duplicates and invalid
        requests as early as possible.

        :param url: URL to follow.
        :param callback: Callback method to run if the Request is successful.
        :param response: Previous Response that contained the Request URL.
        :param kwargs: Optional Request keyword arguments. See Request for details.
        :param method: HTTP method for Request.
        :param delay: Optionally override the default delay for the Request.
        :param priority: Optionally override the default priority of the Request.
        :param allow_domain: Optionally override the allowed domains check.
        :param max_content_length: Optionally override the maximum allowed size in bytes of Response body.
        :param retries: Optionally override the number of Request retries.
        :param timeout: Optionally override the Request timeout.
        :param cb_kwargs: Optional Dictionary of keyword arguments to be passed to the callback function.
        :return: Request
        """
        original_url = copy.copy(url)
        if isinstance(url, str):
            url = parse_href_to_url(url)

        if not url:
            logger.warning("Attempted to follow invalid URL: %s", original_url)
            return

        history = []
        if response:
            # Join the URL to the Response URL if it doesn't contain a domain.
            if not url.is_absolute() or not url.scheme:
                url = coerce_url(
                    response.origin.join(url), default_scheme=response.scheme
                )

            # Restrict the depth of the Request chain to the maximum depth.
            # This test happens before the URL duplicate check so that the URL might still be reachable by another path.
            if self.max_depth and len(response.history) >= self.max_depth:
                logger.debug("Max Depth of '%d' reached: %s", self.max_depth, url)
                return

            # Copy the Response history so that it isn't a reference to a mutable object.
            history = copy.deepcopy(response.history)
        else:
            if not url.is_absolute():
                logger.debug("URL should have domain: %s", url)
                return

            if not url.scheme:
                url = coerce_url(url)

        # The URL scheme must be in the list of allowed schemes.
        if self.allowed_schemes and url.scheme not in self.allowed_schemes:
            logger.debug("URI Scheme '%s' not allowed: %s", url.scheme, url)
            return

        # The URL host must be in the list of allowed domains.
        if not allow_domain and not self.is_allowed_domain(url):
            logger.debug("Domain '%s' not allowed: %s", url.host, url)
            return

        # Check if URL is not already seen, and add it to the duplicate filter seen list.
        if await self._duplicate_filter.url_seen(url, method):
            return

        request = Request(
            url=url,
            request_session=self._session,
            history=history,
            callback=callback,
            xml_parser=self.parse_xml,
            max_content_length=max_content_length or self.max_content_length,
            timeout=timeout or self.request_timeout,
            method=method,
            delay=delay if isinstance(delay, float) else self.delay,
            retries=retries or self.max_retries,
            cb_kwargs=cb_kwargs,
            **kwargs,
        )

        # Override the Request priority only if the kwarg is provided.
        if priority:
            request.priority = priority

        return request

    @abstractmethod
    async def process_item(self, item: Item) -> None:
        """
        Processed a parsed Item in some way. e.g. Add it to the Item set, or database, or send a signal.

        :param item: A parsed Item.
        """
        self.items.add(item)

    @abstractmethod
    async def parse_xml(self, response_text: str) -> Any:
        """
        Parse Response text as XML.
        Used to allow implementations to provide their own XML parser.

        :param response_text: Response text as string.
        """
        raise NotImplementedError("Not Implemented")

    @abstractmethod
    async def parse(self, request: Request, response: Response) -> AsyncGeneratorType:
        """
        Parse an HTTP Response. Must yield Items, Requests, AsyncGenerators, or Coroutines.

        :param request: HTTP Request that created the Response.
        :param response: HTTP Response.
        """
        raise NotImplementedError("Not Implemented")

    def _put_queue(self, queueable: Queueable) -> None:
        """
        Put an object that inherits from Queueable onto the Request Queue.

        :param queueable: An object that inherits from Queueable.
        """
        if not isinstance(queueable, Queueable):
            raise ValueError("Object must inherit from Queueable Class")

        queueable.add_to_queue(self._request_queue)
        self.stats[Stats.QUEUED_TOTAL] += 1

    async def _work(self, task_num):
        """
        Worker function for handling request queue items.
        """
        try:
            while True:
                self._stats_queue_sizes.append(self._request_queue.qsize())
                item: Queueable = await self._request_queue.get()
                # logger.debug("Priority: %s Item: %s", item.priority, item)
                if item.get_queue_wait_time():
                    # logger.debug(
                    #     "Waited: %sms Item: %s", item.get_queue_wait_time(), item
                    # )
                    self._stats_queue_wait_times.append(item.get_queue_wait_time())

                if self._session.closed:
                    logger.debug("Session is closed. Cannot run %s", item)
                    continue

                try:
                    # Fetch Request and handle callbacks
                    if isinstance(item, Request):
                        await self._handle_request(item)
                    # Process Callback results
                    elif isinstance(item, CallbackResult):
                        await self._process_request_callback_result(
                            item.result, item.callback_recursion
                        )
                except Exception as e:
                    logger.exception("Error handling item: %s : %s", item, e)
                finally:
                    self._request_queue.task_done()
        except asyncio.CancelledError:
            logger.debug("Cancelled Worker: %s", task_num)

    @staticmethod
    async def _run_callback(callback, *args, **kwargs) -> None:
        """
        Runs a callback function.

        :param callback: Function to run. May be async.
        :param args: Positional arguments to pass to the function.
        :param kwargs: Keyword arguments to pass to the function.
        :return: None
        """
        if not callback:
            return
        if inspect.iscoroutinefunction(callback):
            await callback(*args, **kwargs)
        elif inspect.isfunction(callback):
            callback(*args, **kwargs)
        else:
            logger.warning("Callback %s must be a coroutine or function", callback)

    def create_start_urls(self, urls: List[Union[URL, str]]) -> List[URL]:
        """
        Create the start URLs for the crawl from an initial URL. May be overridden.

        :param urls: Initial URLs
        """
        crawl_start_urls: Set[URL] = set()

        for url in urls + self.start_urls:
            if isinstance(url, str):
                if "//" not in url:
                    url = f"//{url}"
                url = URL(url)

            if url.scheme.lower() not in ["http", "https"]:
                url = url.with_scheme("http")

            crawl_start_urls.add(url)

        return list(crawl_start_urls)

    def record_statistics(self) -> None:
        """
        Record statistics.
        """
        self.stats[Stats.REQUESTS_DURATION_TOTAL] = int(
            sum(self._stats_request_durations)
        )
        self.stats[Stats.REQUESTS_DURATION_AVG] = int(
            harmonic_mean(self._stats_request_durations)
        )
        self.stats[Stats.REQUESTS_DURATION_MAX] = int(
            max(self._stats_request_durations)
        )
        self.stats[Stats.REQUESTS_DURATION_MIN] = int(
            min(self._stats_request_durations)
        )
        self.stats[Stats.REQUESTS_DURATION_MEDIAN] = int(
            median(self._stats_request_durations)
        )

        self.stats[Stats.CONTENT_LENGTH_TOTAL] = int(
            sum(self._stats_response_content_lengths)
        )
        self.stats[Stats.CONTENT_LENGTH_AVG] = int(
            harmonic_mean(self._stats_response_content_lengths)
        )
        self.stats[Stats.CONTENT_LENGTH_MAX] = int(
            max(self._stats_response_content_lengths)
        )
        self.stats[Stats.CONTENT_LENGTH_MIN] = int(
            min(self._stats_response_content_lengths)
        )
        self.stats[Stats.CONTENT_LENGTH_MEDIAN] = int(
            median(self._stats_response_content_lengths)
        )

        self.stats[Stats.URLS_SEEN] = len(self._duplicate_filter.fingerprints)

        self.stats[Stats.QUEUE_WAIT_AVG] = harmonic_mean(self._stats_queue_wait_times)
        self.stats[Stats.QUEUE_WAIT_MIN] = min(self._stats_queue_wait_times)
        self.stats[Stats.QUEUE_WAIT_MAX] = max(self._stats_queue_wait_times)
        self.stats[Stats.QUEUE_WAIT_MEDIAN] = median(self._stats_queue_wait_times)

        self.stats[Stats.QUEUE_SIZE_MAX] = max(self._stats_queue_sizes)
        self.stats[Stats.QUEUE_SIZE_AVG] = int(harmonic_mean(self._stats_queue_sizes))
        self.stats[Stats.QUEUE_SIZE_MEDIAN] = int(median(self._stats_queue_sizes))

        self.stats[Stats.REQUESTS_LATENCY_AVG] = harmonic_mean(
            self._stats_request_latencies
        )
        self.stats[Stats.REQUESTS_LATENCY_MAX] = int(max(self._stats_request_latencies))
        self.stats[Stats.REQUESTS_LATENCY_MIN] = int(min(self._stats_request_latencies))
        self.stats[Stats.REQUESTS_LATENCY_MEDIAN] = int(
            median(self._stats_request_latencies)
        )
        self.stats[Stats.REQUESTS_LATENCY_TOTAL] = int(
            sum(self._stats_request_latencies)
        )

    def get_stats(self) -> dict:
        """
        Return crawl statistics as a sorted dictionary.
        """
        stats = {str(k): v for k, v in self.stats.items()}
        return dict(OrderedDict(sorted(stats.items())).items())

    async def crawl(self, urls: Union[URL, str, List[Union[URL, str]]] = None) -> None:
        """
        Start the web crawler.

        :param urls: An optional URL or List of URLS to start the crawl, in addition to start_urls.
        """

        # Fix for ssl errors
        ignore_aiohttp_ssl_error(asyncio.get_running_loop())

        start = time.perf_counter()

        # Create start urls from the initial URL if provided.
        if not urls:
            urls = []
        if isinstance(urls, (URL, str)):
            urls = [urls]
        self.start_urls = self.create_start_urls(urls)

        if not self.start_urls:
            raise ValueError("crawler.start_urls are required")

        # Create the Request Queue within the asyncio loop.
        self._request_queue = CrawlerPriorityQueue()

        # Create the Semaphore for controlling HTTP Request concurrency within the asyncio loop.
        self._semaphore = asyncio.Semaphore(self.concurrency)

        trace_configs = []
        if self._trace:
            trace_configs.append(add_trace_config())

        conn = aiohttp.TCPConnector(
            limit=0, ssl=self._ssl, ttl_dns_cache=self.total_timeout.total
        )
        # Create the ClientSession for HTTP Requests within the asyncio loop.
        self._session = aiohttp.ClientSession(
            timeout=self.total_timeout,
            headers=self.headers,
            connector=conn,
            trace_configs=trace_configs,
        )

        # Create a Request for each start URL and add it to the Request Queue.
        for url in self.start_urls:
            req = await self.follow(coerce_url(url), self.parse, delay=0)
            if req:
                self._process_request(req)

        # Create workers to process the Request Queue.
        # Create twice as many workers as potential concurrent requests, to help handle request callbacks without
        # delay while other workers may be locked by the Semaphore.
        self._workers = [
            asyncio.create_task(self._work(i)) for i in range(self.concurrency * 2)
        ]

        try:
            # Run workers within the ClientSession.
            async with self._session:
                await asyncio.wait_for(
                    self._request_queue.join(), timeout=self.total_timeout.total
                )
        except asyncio.TimeoutError:
            logger.debug("Timed out after %s seconds", self.total_timeout.total)
            self._request_queue.clear()
        finally:
            # Make sure all workers are cancelled.
            for w in self._workers:
                w.cancel()
            # Wait until all worker tasks are cancelled.
            await asyncio.gather(*self._workers, return_exceptions=True)

        # Run the post crawl callback if it exists.
        await self._run_callback(self.post_crawl_callback)

        # The ClientSession is closed only after all work is completed.
        await self._session.close()

        duration = int((time.perf_counter() - start) * 1000)
        self.stats[Stats.TOTAL_DURATION] = duration

        self.record_statistics()

        logger.info(
            "Crawl finished: requests=%s time=%dms",
            self.stats[Stats.REQUESTS_QUEUED],
            duration,
        )
        logger.debug("Stats: %s", self.stats)



from asyncio import Queue
from typing import Union

import time


class Queueable:
    queue_put_time = None
    queue_get_time = None
    # Default lowest queue priority is 100 (higher number means lower priority)
    priority = 100

    def get_queue_wait_time(self) -> Union[float, None]:
        """
        Get the time in Milliseconds that this object has been on the queue.

        :return: Queue wait time in Milliseconds as float
        """
        # Only set queue_get_time if not already set, so that the value of this method doesn't change each time
        # it's called.
        if not self.queue_get_time:
            self.queue_get_time = time.perf_counter()
        if self.queue_put_time:
            return (self.queue_get_time - self.queue_put_time) * 1000
        return None

    def set_queue_put_time(self) -> None:
        """
        Set the time that this object was put onto the queue.
        """
        # Set queue_get_time to None, because this method is called whenever a Queueable is added to the queue
        # and it may be added to a queue multiple times in it's life.
        self.queue_get_time = None
        self.queue_put_time = time.perf_counter()

    def add_to_queue(self, queue: Queue) -> None:
        """
        Add the Queueable to the queue and set the queue put time.

        :param queue: An Queue instance
        """
        self.set_queue_put_time()
        queue.put_nowait(self)

    def __lt__(self, other) -> bool:
        """
        Compare Queueable priority for Queue ordering.
        Lower priority has precedence in the Queue.

        :param other: Another Queueable object
        :return: boolean
        """
        if not isinstance(other, Queueable):
            return True
        return self.priority < other.priority



import logging
from datetime import datetime, date
from statistics import mean
from types import AsyncGeneratorType
from typing import Tuple, List, Union, Dict

import feedparser
import time
from aiohttp import hdrs
from bs4 import BeautifulSoup
from yarl import URL

from feedsearch_crawler.crawler import ItemParser, Request, Response, to_string
from feedsearch_crawler.crawler.lib import headers_to_dict, remove_www
from feedsearch_crawler.feed_spider.favicon import Favicon
from feedsearch_crawler.feed_spider.feed_info import FeedInfo
from feedsearch_crawler.feed_spider.lib import (
    parse_header_links,
    datestring_to_utc_datetime,
    create_content_type,
    ParseTypes,
)

logger = logging.getLogger(__name__)


class FeedInfoParser(ItemParser):
    async def parse_item(
        self, request: Request, response: Response, *args, **kwargs
    ) -> AsyncGeneratorType:
        logger.info("Parsing: Feed %s", response.url)

        if "parse_type" not in kwargs:
            raise ValueError("type keyword argument is required")

        parse_type = kwargs["parse_type"]

        content_type = create_content_type(
            parse_type,
            response.encoding,
            response.headers.get(hdrs.CONTENT_TYPE, "").lower(),
        )

        item = FeedInfo(url=response.url, content_type=content_type)

        # Check link headers first for WebSub content discovery
        # https://www.w3.org/TR/websub/#discovery
        if response.headers:
            item.hubs, item.self_url = self.header_links(response.headers)

        try:
            valid_feed = False

            if parse_type == ParseTypes.JSON:
                valid_feed = self.parse_json(item, response.json)
            elif parse_type == ParseTypes.XML:
                valid_feed = self.parse_xml(
                    item,
                    response.data,
                    response.encoding,
                    headers_to_dict(response.headers),
                )

            if not valid_feed:
                logger.debug("Invalid Feed: %s", item)
                return
        except Exception as e:
            logger.exception("Failed to parse feed %s, Error: %s", item, e)
            return

        if item.favicon and self.crawler.favicon_data_uri:
            favicon = Favicon(
                url=item.favicon,
                priority=1,
            )
            yield self.follow(
                item.favicon,
                self.crawler.parse_favicon_data_uri,
                cb_kwargs=dict(favicon=favicon),
            )

        self.validate_self_url(item)

        item.content_length = response.content_length
        self.score_item(item, response.history[0])
        yield item

    def parse_xml(
        self, item: FeedInfo, data: Union[str, bytes], encoding: str, headers: Dict
    ) -> bool:
        """
        Get info from XML (RSS or ATOM) feed.
        """

        # Parse data with feedparser
        try:
            parsed = self.parse_raw_data(data, encoding, headers)
        except Exception as e:
            logger.exception("Unable to parse feed %s: %s", item, e)
            return False

        if not parsed:
            logger.warning("No valid feed data for %s", item)
            return False

        if parsed.get("bozo") == 1:
            bozo_exception = parsed.get("bozo_exception", None)
            if isinstance(bozo_exception, feedparser.CharacterEncodingOverride):
                item.bozo = 1
            elif isinstance(
                bozo_exception,
                (feedparser.CharacterEncodingUnknown, feedparser.UndeclaredNamespace),
            ):
                logger.warning("No valid feed data for %s: %s", item, bozo_exception)
                return False

        feed = parsed.get("feed")
        if not feed:
            return False
        if not parsed.get("entries"):
            return False

        # Only search if no hubs already present from headers
        if not item.hubs:
            item.hubs, item.self_url = self.websub_links(feed)

        if item.hubs and item.self_url:
            item.is_push = True

        item.version = parsed.get("version")
        item.title = self.feed_title(feed)
        item.description = self.feed_description(feed)
        item.is_podcast = self.is_podcast(parsed)

        try:
            dates = []
            now_date = datetime.utcnow().date()

            entries = parsed.get("entries", [])
            item.item_count = len(entries)

            dates.extend(
                FeedInfoParser.entry_dates(entries, ["updated", "published"], now_date)
            )

            if dates:
                item.last_updated = sorted(dates, reverse=True)[0]
                item.velocity = self.entry_velocity(dates)
            elif feed.get("updated"):
                item.last_updated = datestring_to_utc_datetime(feed.get("updated"))
        except Exception as e:
            logger.exception("Unable to get feed published date: %s", e)
            pass

        return True

    def parse_json(self, item: FeedInfo, data: dict) -> bool:
        """
        Get info from JSON feed.

        :param item: FeedInfo object
        :param data: JSON object
        :return: None
        """
        item.version = data.get("version")
        if "https://jsonfeed.org/version/" not in item.version:
            item.bozo = 1
            return False

        if not data.get("items"):
            return False

        item.title = data.get("title")
        item.description = data.get("description")

        favicon = data.get("favicon")
        if favicon:
            item.favicon = URL(favicon)

        # Only search if no hubs already present from headers
        if not item.hubs:
            try:
                item.hubs = list(hub.get("url") for hub in data.get("hubs", []))
            except (IndexError, AttributeError):
                pass

        if item.hubs:
            item.is_push = True

        try:
            dates = []
            now_date: date = datetime.utcnow().date()

            entries = data.get("items", [])
            item.item_count = len(entries)

            dates.extend(
                FeedInfoParser.entry_dates(
                    entries, ["date_modified", "date_published"], now_date
                )
            )

            if dates:
                item.last_updated = sorted(dates, reverse=True)[0]
                item.velocity = self.entry_velocity(dates)
        except Exception as e:
            logger.exception("Unable to get feed published date: %s", e)
            pass

        return True

    @staticmethod
    def parse_raw_data(
        raw_data: Union[str, bytes], encoding: str = "utf-8", headers: Dict = None
    ) -> Dict:
        """
        Loads the raw RSS/Atom XML data.
        Returns feedparser Dict.
        https://pythonhosted.org/feedparser/

        :param raw_data: RSS/Atom XML feed
        :type raw_data: str
        :param encoding: Character encoding of raw_data
        :type encoding: str
        :param headers: Response headers
        :return: Dict
        """
        if not encoding:
            encoding = "utf-8"

        h = {}
        if headers:
            if isinstance(headers, dict):
                h = headers
            else:
                try:
                    h.update({k.lower(): v for (k, v) in headers.items()})
                except KeyError:
                    pass

            h.pop("content-encoding", None)

        try:
            start = time.perf_counter()

            if isinstance(raw_data, str):
                raw_data: bytes = raw_data.encode(encoding)

            raw_data = raw_data.strip()
            content_length = len(raw_data)

            # We want to pass data into feedparser as bytes, otherwise if we accidentally pass a url string
            # it will attempt a fetch
            data = feedparser.parse(raw_data, response_headers=h)

            dur = int((time.perf_counter() - start) * 1000)
            logger.debug("Feed Parse: size=%s dur=%sms", content_length, dur)

            return data
        except Exception as e:
            logger.exception("Could not parse RSS data: %s", e)

    def feed_title(self, feed: dict) -> str:
        """
        Get feed title

        :param feed: feed dict
        :return: str
        """
        title = feed.get("title", None)
        if not title:
            return ""
        return self.clean_title(title)

    def clean_title(self, title: str) -> str:
        """
        Cleans title string, and shortens if too long.
        Have had issues with dodgy feed titles.

        :param title: Title string
        :return: str
        """
        try:
            title = BeautifulSoup(title, self.crawler.htmlparser).get_text()
            if len(title) > 1024:
                title = title[:1020] + "..."
            return title
        except Exception as ex:
            logger.exception("Failed to clean title: %s", ex)
            return ""

    @staticmethod
    def is_podcast(parsed: dict) -> bool:
        """
        Check if the feed is a Podcast.

        :param parsed: Feedparser dict
        :return: bool
        """
        if not parsed:
            return False

        has_itunes: bool = "itunes" in parsed.get("namespaces", {})
        has_enclosures: bool = any(
            [len(x.get("enclosures", [])) >= 1 for x in parsed.get("entries", [])]
        )
        return has_itunes and has_enclosures

    @staticmethod
    def feed_description(feed: dict) -> str:
        """
        Get feed description.

        :param feed: feed dict
        :return: str
        """
        subtitle = feed.get("subtitle", None)
        if subtitle:
            return subtitle
        return feed.get("description", None)

    @staticmethod
    def websub_links(feed: dict) -> Tuple[List[str], str]:
        """
        Returns a tuple containing the hub url and the self url for
        a parsed feed.

        :param feed: An RSS feed parsed by feedparser
        :type feed: dict
        :return: tuple
        """
        links = feed.get("links", [])
        return FeedInfoParser.find_hubs_and_self_links(links)

    @staticmethod
    def header_links(headers: dict) -> Tuple[List[str], str]:
        """
        Attempt to get self and hub links from HTTP headers
        https://www.w3.org/TR/websub/#x4-discovery

        :param headers: Dict of HTTP headers
        :return: None
        """
        link_header = headers.get("Link")
        links: list = []

        if link_header:
            links = parse_header_links(to_string(link_header))

        return FeedInfoParser.find_hubs_and_self_links(links)

    @staticmethod
    def find_hubs_and_self_links(links: List[dict]) -> Tuple[List[str], str]:
        """
        Parses a list of links into self and hubs urls

        :param links: List of parsed HTTP Link Dicts
        :return: Tuple
        """
        hub_urls: List[str] = []
        self_url: str = ""

        if not links:
            return [], ""

        for link in links:
            try:
                if link["rel"] == "hub":
                    href: str = link["href"]
                    hub_urls.append(href)
                elif link["rel"] == "self":
                    self_url = link["href"]
            except KeyError:
                continue

        return hub_urls, self_url

    @staticmethod
    def score_item(item: FeedInfo, original_url: URL):
        score = 0

        url_str = str(item.url).lower()

        # -- Score Decrement --

        if original_url:
            host = remove_www(original_url.host)

            if host not in item.url.host:
                score -= 20

        # Decrement the score by every extra path in the url
        parts_len = len(item.url.parts)
        if parts_len > 2:
            score -= (parts_len - 2) * 2

        if item.bozo:
            score -= 20
        if not item.description:
            score -= 10
        if "georss" in url_str:
            score -= 10
        if "alt" in url_str:
            score -= 7
        if "comments" in url_str or "comments" in item.title.lower():
            score -= 15
        if "feedburner" in url_str:
            score -= 10

        # -- Score Increment --
        if item.url.scheme == "https":
            score += 10
        if item.is_push:
            score += 10
        if "index" in url_str:
            score += 30

        if "comments" in url_str or "comments" in item.title.lower():
            score -= 15
        else:
            score += int(item.velocity)

        if any(map(url_str.count, ["/home", "/top", "/most", "/magazine"])):
            score += 10

        kw = ["atom", "rss", ".xml", "feed", "rdf"]
        for p, t in zip(range(len(kw) * 2, 0, -2), kw):
            if t in url_str:
                score += p

        item.score = score

    @staticmethod
    def entry_dates(entries: List[Dict], date_names: List[str], current_date: date):
        """
        Return published or updated dates from feed entries.

        :param entries: List of feed entries as dicts.
        :param date_names: List of key names of entry published or updated values.
        :param current_date: The current date.
        :return: generator that returns datetimes.
        """
        for entry in entries:
            for name in date_names:
                try:
                    entry_date: datetime = datestring_to_utc_datetime(entry[name])
                    if entry_date.date() <= current_date:
                        yield entry_date
                except (KeyError, ValueError):
                    pass

    @staticmethod
    def entry_velocity(dates: List[datetime]) -> float:
        """
        Calculate velocity of posted entries, returns a float of the average number of entries posted per day.

        :param dates: List of entry dates
        :return: Average entries per day
        """
        if not dates or len(dates) < 3:
            return 0

        dates = sorted(dates)
        deltas = []
        previous_date: datetime = dates[0]

        for current_date in dates[1:]:
            if current_date == previous_date:
                continue
            delta = current_date - previous_date
            deltas.append(delta.total_seconds())
            previous_date = current_date

        if not deltas:
            return 0

        mean_seconds_delta = mean(deltas)

        result = round(86400 / mean_seconds_delta, 3)
        return result

    @staticmethod
    def validate_self_url(item: FeedInfo) -> None:
        """
        Validate the self url

        :param item: FeedInfo item
        """
        try:
            item.self_url = URL(item.self_url)
        except ValueError:
            item.self_url = ""
            return

        if item.self_url and item.self_url != item.url:
            # Handle a case where the item url contains a trailing slash and the self url doesn't.
            if str(item.url).strip("/") == str(item.self_url):
                item.url = URL(str(item.url).strip("/"))
                return

            # The self url should be an absolute url.
            if not item.self_url.is_absolute():
                if str(item.self_url) in str(item.url):
                    item.self_url = item.url
                else:
                    item.self_url = ""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
from codecs import open

from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, "Readme.md"), encoding="utf-8") as f:
    readme = f.read()


def build():
    os.system("python3 setup.py sdist bdist_wheel")


def publish():
    os.system("twine upload dist/*")


if sys.argv[-1] == "build":
    build()
    sys.exit()
elif sys.argv[-1] == "publish":
    build()
    publish()
    sys.exit()
elif sys.argv[-1] == "publish-only":
    publish()
    sys.exit()


packages = [
    "feedsearch_crawler",
    "feedsearch_crawler.crawler",
    "feedsearch_crawler.feed_spider",
]

required = [
    "aiohttp",
    "beautifulsoup4",
    "feedparser",
    "cchardet",
    "aiodns",
    "w3lib",
    "uvloop ; sys_platform != 'win32'",
]

setup(
    name="feedsearch-crawler",
    version="1.0.0",
    description="Search sites for RSS, Atom, and JSON feeds",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="David Beath",
    author_email="davidgbeath@gmail.com",
    url="https://github.com/DBeath/feedsearch-crawler",
    license="MIT",
    packages=packages,
    install_requires=required,
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Development Status :: 5 - Production/Stable",
        "Natural Language :: English",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Internet :: WWW/HTTP :: Indexing/Search",
        "Typing :: Typed",
        "Framework :: AsyncIO",
    ],
    python_requires=">=3.7",
)



from feedsearch_crawler.crawler.crawler import Crawler
from feedsearch_crawler.crawler.duplicatefilter import DuplicateFilter
from feedsearch_crawler.crawler.item import Item
from feedsearch_crawler.crawler.item_parser import ItemParser
from feedsearch_crawler.crawler.lib import (
    to_string,
    to_bytes,
    coerce_url,
    CallbackResult,
)
from feedsearch_crawler.crawler.request import Request
from feedsearch_crawler.crawler.response import Response

__all__ = [
    "Crawler",
    "Item",
    "ItemParser",
    "DuplicateFilter",
    "Request",
    "Response",
    "to_bytes",
    "to_string",
    "coerce_url",
    "CallbackResult",
]
